//********************************************************************************
//*
//*		SCOMM - Serial Bit Banged Output Interface
//*
//*	2014 Nov   :	V1 / bhilpert
//*			Using busy-wait delay for bits (clock() function) produced delays too variable and unpredictable.
//*			V2
//*			Changed to SIGALRM technique. Using it_interval to restart the alarm automatically doesn't work,
//*			so setitimer is called for each bit. 
//*
//********************************************************************************
#include <sys/types.h>
#include <sys/time.h>
#include <signal.h>
#include <unistd.h>
#include <stdio.h>
#ifdef __linux__
	#include <wiringPi.h>
#else
	#include "../util/gpioDummy.h"
#endif
#include "util.h"


//----------------------------------------------
// wiringPi pin numbers
// Note these are different than both the header pin numbers and the BCM GPIO numbers

#define PIN_SDO			12						// header pin 19; MOSI


//----------------------------------------------
// Interface config & state descriptor

typedef struct
 {	int			uSBit;						// bit period in uS
	int			uSStop;						// stop bits period in uS
	int			dataBits;					// number of data bits

	int			bits;						// bits being sent, composed of stop, data & start
	struct itimerval	bitTm;						// interval for bit period
 }
	CPSBB;

private CPSBB			_sbb;						// HACK: global


//----------------------------------------------
// Alarm handler to send next bit at each bit period

private void
_cpSBB_SendBit( sig )
	int			sig;
 {
	if( (_sbb.bits>>=1) == 0 )						// shift bits down till stop bit disappears, when it does byte has been sent 
		return;

	if( _sbb.bits == 1 )							// when bits hits 1 it's the stop bit, adjust final bit period for number of stop bits
		_sbb.bitTm.it_value.tv_usec = _sbb.uSStop;

	setitimer( ITIMER_REAL, &_sbb.bitTm, NULL );				// restart timer. Should be using it_interval but it doesn't work !!!

	digitalWrite( PIN_SDO, _sbb.bits&1 );					// send bit
 }


//* END Private
//********************************************************************************
//* Public

//-----------------------------------------------
// Open interface

void*
cpSBB_open( name, speed, bits )
	char			*name;
	int			speed, bits;
 {
	CPSBB			*sbb = &_sbb;
	int			stopHalves;

	if( speed < 0 )
		sbb->uSBit = 1000000*100 / -speed;				// if speed is negative it is in hundredths, that is, 100 times the intended rate
	else
		sbb->uSBit = 1000000 / speed;					// calculate bit period in microseconds

	switch( bits )								// set bits/char
	 {	case 5:		sbb->dataBits=5;	stopHalves=2;	break;	// 1 stop bit
		case 6:		sbb->dataBits=6;	stopHalves=2;	break;
		case 7:		sbb->dataBits=7;	stopHalves=2;	break;
		case 8:		sbb->dataBits=8;	stopHalves=2;	break;

		case 52:	sbb->dataBits=5;	stopHalves=4;	break;	// 2 stop bits
		case 62:	sbb->dataBits=6;	stopHalves=4;	break;
		case 72:	sbb->dataBits=7;	stopHalves=4;	break;
		case 82:	sbb->dataBits=8;	stopHalves=4;	break;

		case 515:	sbb->dataBits=5;	stopHalves=3;	break;	// 1.5 stop bits

		default:
			fprintf( stderr, "  -- invalid bits/char spec\n" );
			return( NULL );
	 }
	sbb->uSStop = (sbb->uSBit * stopHalves) >> 1;				// calculate stop-bits period

	wiringPiSetup();							// initialize GPIO library
	pinMode( PIN_SDO, OUTPUT );						// config serial data output

	digitalWrite( PIN_SDO, 1 );						// idle in 1/mark state

	sbb->bits = 0;								// set state for no byte being sent			
	signal( SIGALRM, _cpSBB_SendBit );					// register handler to send bits at each bit period

	return( sbb );
 }


//-----------------------------------------------
// Close interface

void
cpSBB_close( sbb )
	CPSBB*			sbb;
 {
 }


//-----------------------------------------------
// Read a byte, dummy for now

int
cpSBB_rByte( sbb )
	CPSBB*			sbb;
 {
	repeat
	 {	sleep( 1000000 );
	 }

	return( 0 );
 }


//-----------------------------------------------
// Write a byte

void
cpSBB_wByte( sbb, byte )
	CPSBB*			sbb;
	int			byte;
 {
	while( sbb->bits ) sleep( 1 );						// wait for previous byte to be sent, timer signal interrupts sleep()

	byte &= ((1<<sbb->dataBits)-1);						// mask data to ensure bits at top are clear
	sbb->bits = ((1<<sbb->dataBits) | byte) << (1+1);			// set stop bit at top to 1/mark state, shift bits up to make room for start bit,
										//	and to compensate for extra shift in first SendBit
	sbb->bitTm.it_interval.tv_usec	= 0; //sbb->uSBit;			// set alarm restart interval to bit periods
	sbb->bitTm.it_interval.tv_sec	= 0;
	sbb->bitTm.it_value.tv_usec	= sbb->uSBit;				// set first interval to bit period
	sbb->bitTm.it_value.tv_sec	= 0;

	_cpSBB_SendBit( 0 );							// send first bit
 }


//* EOF
//********************************************************************************
