//***************************************************************************************************************************
//*
//*			SCOMM - Serial Communication Port Interface for UNIX TTY device
//*
//* 2014 Oct   : V1 / bhilpert
//*
//* Some devices may block on open waiting for line detection such as DCD, which may be undesirable.
//* To get around this, the device is opened in non-blocking mode to get a filedes, blocking on r/w then re-enabled and line detection disabled.
//*
//***************************************************************************************************************************
#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <termios.h>
#include <errno.h>
#include "util.h"


//----------------------------------------------
// Interface config & state descriptor

typedef struct
 {	int			fno;
 }
	CPTTY;

private CPTTY			_tty;						// HACK: global


//* END Private
//***************************************************************************************************************************
//* Public

//----------------------------------------------
// Open interface

void*
cpTTY_open( devName, speed, bits )
	char*			devName;
	int			speed, bits;
 {
	struct termios		tattr;
	int			fno, flags;

	//--------------------------------
	// Open the device

	fno = open( devName, O_RDWR|O_NONBLOCK );				// open without blocking

	if( fno < 0 )
	 {	fprintf( stderr, "  -- failed to open serial port\n" );
		goto err0;
	 }

	if( !isatty(fno) )
	 {	fprintf( stderr, "  -- specified device is not a serial port\n" );
		goto err1;
	 }

	//--------------------------------
	// Set IO flags to re-enable blocking on r/w

	flags = fcntl( fno, F_GETFL, 0 );
	if( flags == -1 )
	 {	fprintf( stderr, "  -- error on GETFL\n" );
		goto err1;
 	 }
	flags &= ~O_NONBLOCK;
	if( fcntl(fno,F_SETFL,flags) == -1 )
	 {	fprintf( stderr, "  -- error on SETFL %d %s\n", fno, devName );
		goto err1;
 	 }

	//--------------------------------
	// Set TTY device attributes

	tcgetattr( fno, &tattr );						// get attributes to modify

	if( cfsetspeed(&tattr,speed) != 0 )					// assign speed
	 {	fprintf( stderr, "  -- failed to assign speed\n" );
		goto err1;
	 }

	tattr.c_cflag &= ~(CSIZE|CSTOPB);					// assign bits/char and stop bits
	switch( bits )
	 {	case 5:  tattr.c_cflag |= CS5;		break;
		case 6:  tattr.c_cflag |= CS6;		break;
		case 7:  tattr.c_cflag |= CS7;		break;
		case 8:  tattr.c_cflag |= CS8;		break;

		case 52: tattr.c_cflag |= CS5 | CSTOPB;	break;
		case 62: tattr.c_cflag |= CS6 | CSTOPB;	break;
		case 72: tattr.c_cflag |= CS7 | CSTOPB;	break;
		case 82: tattr.c_cflag |= CS8 | CSTOPB;	break;

		default:
			fprintf( stderr, "  -- invalid bits/char spec\n" );
			goto err1;
	 }

	tattr.c_cflag &= ~PARENB;						// disable parity
	tattr.c_cflag |= CLOCAL;						// disable potential line detection requirement

	tattr.c_lflag &= ~(ICANON|ECHO);					// turn off input canonical processing and echo
	tattr.c_cc[VMIN]  = 1;
	tattr.c_cc[VTIME] = 0;

	tattr.c_iflag &= ~(IGNCR|ICRNL|INLCR);					// pass CR & NL through
	tattr.c_iflag &= ~(IXOFF|IXON);						// no flow control on input or output

	tattr.c_oflag &= ~OPOST;						// turn off output character processing

	if( tcsetattr(fno,TCSAFLUSH,&tattr) < 0 )				// set new attributes
	 {	fprintf( stderr, "  -- error setting comm port attributes\n" );
	 err1:	close( fno );
	 err0:	return( NULL );
	 }

	//--------------------------------
	// Init descriptor

	_tty.fno = fno;

	return( &_tty );
 }


//----------------------------------------------
// Close interface

void
cpTTY_close( tty )
	CPTTY*			tty;
 {
	close( tty->fno );
 }


//----------------------------------------------
// Read a byte

int
cpTTY_rByte( tty )
	CPTTY*			tty;
 {
	char			b[1];

	if( read(tty->fno,b,1) <= 0 )
		return( errno==EINTR ? -2 : -1 );

	return( b[0] & 0xFF );
 }


//----------------------------------------------
// Write a byte

void
cpTTY_wByte( tty, byte )
	CPTTY*			tty;
	int			byte;
 {
	char			b[1];

	b[0] = byte;
	write( tty->fno, b, 1 );
 }


//** EOF
//***************************************************************************************************************************
