//***************************************************************************************************************************
//*
//*			SCOMM - Receiver Process
//*
//* 2014 Oct	: V1 / bhilpert
//*			- Directive signal handler changes config and state underneath main process.
//*			Should be done synchronously. (Fixed)
//*			- Also relies on getting one signal per directive.
//*			If multiple directives get queued, handler will get behind. (Fixed)
//* 2014 Nov	: Changed to interrupt read primitive, so directive can be executed synchronously,
//*		and non-blocking on pipe read so it can loop till directives exhausted.
//*
//***************************************************************************************************************************
#include <unistd.h>
#include <errno.h>
#include <stdio.h>
#include <fcntl.h>
#include <signal.h>
#include "util.h"
#include "scp.h"


//----------------------------------
// Receiver config & state

private SCP*			rcvr__commPort;
private int			rcvr__rPipe;					// read end of directive pipe from sender

private int			rcvr__run;
private FILE*			rcvr__collectFd;				// file to collect read data in
private int			rcvr__doDisplay;				// send received data to display
private int			rcvr__dhex;					// display hex/bin control 0=bin, 1=undefs as hex, 2=all ctl as hex
private int			rcvr__dnl;					// New-Line translation for display
private int			rcvr__fnl;					// New-Line translation for file
private int			rcvr__prevByte;					// most recently received character


//----------------------------------
// Receiver signal handler, dummy just to activate SIGIO

private void
rcvr__onSig( sig )
	int			sig;
 {
 }


//----------------------------------
// Receiver byte filter, optional New-Line translation and non-printables converted to hex

private void
rcvr__put( fd, b, pb, doHex, doNL )
	FILE*			fd;
	int			b, pb;
	int			doHex, doNL;
 {
	if( b>=' ' && b<='~' )							// simple text characters go straight thru
	 {	fputc( b, fd );
		return;
	 }

	if( b=='\t' || b=='\0' )						// tab and null may get additional hex output
	 {	if( doHex == 2 ) fprintf( fd, "\\x%02X", b );
		fputc( b, fd );
		return;
	 }

	if( b=='\r' || b=='\n' )						// cr & lf get special treatment
	 {	if( doHex == 2 ) fprintf( fd, "\\x%02X", b );
		if( doNL )							// check for CR->NL translation
		 {	if( pb=='\r' && b=='\n' )				// if CRLF sequence, NL was already done
				return;
			b = '\n';
		 }
		fputc( b, fd );
		return;
	 }

	if( doHex )								// undefined characters as hex or bin
		fprintf( fd, "\\x%02X", b );
	else
		fputc( b, fd );
 }


//----------------------------------
// Receiver directive processing
// Get directives from parent process through pipe
//
// Pipe directive syntax: '\x01'<cmd><option><arg>*<null>

private void
rcvr__directive()
 {
	char			cmd[100];
	int			n;

	repeat
	 {	if( (n=read(rcvr__rPipe,cmd,100)) != 100 )			// get directive from sender process through pipe
		 {	if( n > 0 )
				goto badDir;
			if( n == 0 )						// EOF on pipe, sender is quitting so quit receiver
			 {	rcvr__run = FALSE;
				break;
			 }
			if( errno == EINTR )					// another signal interrupted, go read again
				continue;
			if( errno != EAGAIN )					// EAGAIN is normal exit from directive processing
				fprintf( stderr, "[SCOMR: bad directive pipe read %d]\n", errno );
			break;
		 }
		if( cmd[0] != '\x01' )						// confirm beginning of directive
			goto badDir;

		//fprintf( stderr, "[SCOMR dir: %s]\n", cmd );
		switch( cmd[1] )
		 {	case 'h':						// DIRECTIVE: set hex/bin mode
				switch( cmd[2] )
				 {	case 'd': rcvr__dhex = cmd[3]-'0';	break;
					default: goto badDir;
				 }
				break;
	
			case 'l':						// DIRECTIVE: set cr->nl mode
				switch( cmd[2] )
				 {	case 'd': rcvr__dnl = cmd[3]!='0';	break;
					case 'f': rcvr__fnl = cmd[3]!='0';	break;
					default: goto badDir;
				 }
				break;

			case 'f':						// DIRECTIVE: collect into file
				if( cmd[2] == 'e' )				// end collecting
				 {	if( !rcvr__collectFd )
						break;

					fclose( rcvr__collectFd );
					rcvr__collectFd = NULL;
					rcvr__doDisplay = TRUE;

					fprintf( stderr, "[SCOMR: collecting terminated]\n" );
					break;
				 }

				if( rcvr__collectFd )
				 {	fprintf( stderr, "[SCOMR: already collecting]\n" );
					break;
				 }

				rcvr__collectFd = fopen( &cmd[3], "w" );	// open collection file
				if( !rcvr__collectFd )
				 {	fprintf( stderr, "[SCOMR: failed to open collection file]\n" );
					break;
				 }

				if( cmd[2] == 's' ) rcvr__doDisplay = FALSE;

				fprintf( stderr, "[SCOMR: collecting to file %s]\n", &cmd[3] );
				break;

			case 'c':						// DIRECTIVE: code translation
				scp_code( rcvr__commPort, &cmd[2] );
				break;

			case 's':						// DIRECTIVE: display status
				fprintf( stderr, "[SCOMR: collecting=%s, hex=%s, rdnl=%s, rfnl=%s, code=%s]\n",
					rcvr__collectFd	? "on" : "off",
					rcvr__dhex==0	? "off" : (rcvr__dhex==1?"undef":"all"),
					rcvr__dnl	? "on" : "off",
					rcvr__fnl	? "on" : "off",
					scp_code(rcvr__commPort,NULL) );
				break;

			default:
		 badDir:	cmd[100-1] = 0;					// ensure there is a null term
				fprintf( stderr, "[SCOMR: bad directive %s]\n", cmd );
				break;
		 }
	 }
 }


//* END Private
//***************************************************************************************************************************
//* Public

//----------------------------------
// Receiver Process

void
Receiver( commPort, rPipe )
	SCP*			commPort;
	int			rPipe;
 {
	struct sigaction	sigAct;
	int			flags, b;

	//---------------------------------------
	// Init

	rcvr__commPort	= commPort;
	rcvr__rPipe	= rPipe;

	rcvr__run	= TRUE;
	rcvr__dhex	= TRUE;
	rcvr__dnl	= FALSE;
	rcvr__fnl	= FALSE;
	rcvr__doDisplay	= TRUE;
	rcvr__collectFd	= NULL;
	rcvr__prevByte	= -1;

	//--------------------------------
	// Prepare pipe & signalling for directives from sender

	if( fcntl(rcvr__rPipe,F_SETOWN,getpid()) == -1 )			// pipe SIGIO signals shall come to this process
	 {	fprintf( stderr, "[SCOMR: error on SETOWN]\n" );
	 	goto fin0;
 	 }

	flags = fcntl( rcvr__rPipe, F_GETFL, 0 );				// get, modify and set pipe descriptor flags
	if( flags == -1 )
	 {	fprintf( stderr, "[SCOMR: error on GETFL]\n" );
		goto fin0;
 	 }
	flags |= O_ASYNC | O_NONBLOCK;						// ASYNC so we get SIGIO, NONBLOCK so we can loop till directives exhausted
	if( fcntl(rcvr__rPipe,F_SETFL,flags) == -1 )
	 {	fprintf( stderr, "[SCOMR: error on SETFL]\n" );
	 	goto fin0;
 	 }

	sigaction( SIGIO, NULL, &sigAct );
	sigAct.sa_handler = &rcvr__onSig;
	sigAct.sa_flags &= ~SA_RESTART;						// allow sys primitives (read) to be interrupted, to process directives
	sigaction( SIGIO, &sigAct, NULL );

	//---------------------------------------
	// Activity loop

	fprintf( stderr, "[SCOMR: ready]\n" );

	while( rcvr__run )
	 {	b = scp_rByte( commPort );					// transfer bytes from comm port to display and/or collection file
		if( b < 0 )
		 {	if( b != -2 )						// -2 indicates interrupted for directive, otherwise EOF or error, get out
				break;
			rcvr__directive();
			continue;
		 }

		if( rcvr__doDisplay )						// output to display
		 {	rcvr__put( stdout,          b, rcvr__prevByte, rcvr__dhex, rcvr__dnl );
			fflush( stdout );
		 }
		if( rcvr__collectFd )						// output to file
			rcvr__put( rcvr__collectFd, b, rcvr__prevByte, FALSE,  rcvr__fnl );

		rcvr__prevByte = b;
	 }

	//---------------------------------------
	// Terminate

 fin0:	fprintf( stderr, "[SCOMR: terminating]\n" );
 }


//** EOF
//***************************************************************************************************************************
